# SPDX-FileCopyrightText: Red Hat, Inc.
# SPDX-License-Identifier: GPL-2.0-or-later

"""
_internal - imageio internal moudules used by the server and client.

Do not use modules form this package; they are internal implemention details
and may change on each release in incompatible way.
"""
